@@ -21,10 +21,9 @@ static void log_data(unsigned char *data, int len)
 		sleep(1);
 		hlaseno = 1;
 	}
-	EINTRLOOP(fd, open(LOG_TRANSFER, O_WRONLY | O_APPEND | O_CREAT, 0600));
+	fd = c_open3(cast_uchar LOG_TRANSFER, O_WRONLY | O_APPEND | O_CREAT, 0600);
 	if (fd != -1) {
 		int rw;
-		set_bin(fd);
 		EINTRLOOP(rw, write(fd, data, len));
 		EINTRLOOP(rw, close(fd));
 	}
@@ -34,35 +33,45 @@ static void log_data(unsigned char *data, int len)
 #define log_data(x, y)
 #endif
 
-#ifdef LOG_SSL
+#if defined(LOG_SSL) && defined(HAVE_OPENSSL)
 #include <openssl/err.h>
 #endif
 
+#ifdef HAVE_SSL
 static inline void log_ssl_error(int line, int ret1, int ret2)
 {
 #ifdef LOG_SSL
 	SSL_load_error_strings();
+#ifdef HAVE_OPENSSL
 	ERR_print_errors_fp(stderr);
+#else
+	{
+		unsigned long err;
+		while ((err = ERR_get_error())) {
+			unsigned char buf[1024];
+			ERR_error_string_n(err, cast_char buf, sizeof buf);
+			fprintf(stderr, "%s\n", buf);
+		}
+	}
+#endif
 	debug("ssl error at %d: %d, %d, %d (%s)", line, ret1, ret2, errno, strerror(errno));
 #endif
 }
+#endif
 
 static void connected(struct connection *);
+static void update_dns_priority(struct connection *);
+static void connected_callback(struct connection *);
 static void dns_found(struct connection *, int);
+static void retry_connect(struct connection *, int, int);
 static void try_connect(struct connection *);
 static void handle_socks_reply(struct connection *);
 
-static void exception(struct connection *c)
-{
-	setcstate(c, S_EXCEPT);
-	retry_connection(c);
-}
-
 int socket_and_bind(int pf, unsigned char *address)
 {
 	int s;
 	int rs;
-	EINTRLOOP(s, socket(pf, SOCK_STREAM, IPPROTO_TCP));
+	s = c_socket(pf, SOCK_STREAM, IPPROTO_TCP);
 	if (s == -1)
 		return -1;
 	if (address && *address) {
@@ -129,8 +138,8 @@ void close_socket(int *s)
 {
 	int rs;
 	if (*s == -1) return;
+	set_handlers(*s, NULL, NULL, NULL);
 	EINTRLOOP(rs, close(*s));
-	set_handlers(*s, NULL, NULL, NULL, NULL);
 	*s = -1;
 }
 
@@ -138,6 +147,7 @@ struct conn_info {
 	void (*func)(struct connection *);
 	struct lookup_result addr;
 	int addr_index;
+	int dont_try_more_servers;
 	int first_error;
 	int port;
 	int *sock;
@@ -196,6 +206,14 @@ void make_connection(struct connection *c, int port, int *sock, void (*func)(str
 	c->newconn = b;
 	log_data(cast_uchar "\nCONNECTION: ", 13);
 	log_data(host, strlen(cast_const_char host));
+#ifdef LOG_TRANSFER
+	{
+		char n[64];
+		snprintf(cast_char n, sizeof n, "%d", port);
+		log_data(":", 1);
+		log_data(n, strlen(cast_const_char n));
+	}
+#endif
 	log_data(cast_uchar "\n", 1);
 	if (c->no_cache >= NC_RELOAD) as = find_host_no_cache(host, &b->addr, &c->dnsquery, (void(*)(void *, int))dns_found, c);
 	else as = find_host(host, &b->addr, &c->dnsquery, (void(*)(void *, int))dns_found, c);
@@ -237,7 +255,7 @@ int get_pasv_socket(struct connection *c, int cc, int *sock, unsigned char *port
 		errno = EINVAL;
 		goto e;
 	}
-	EINTRLOOP(s, socket(PF_INET, SOCK_STREAM, IPPROTO_TCP));
+	s = c_socket(PF_INET, SOCK_STREAM, IPPROTO_TCP);
 	if (s == -1) goto e;
 	*sock = s;
 	set_nonblock(s);
@@ -261,7 +279,6 @@ int get_pasv_socket(struct connection *c, int cc, int *sock, unsigned char *port
 }
 
 #ifdef SUPPORT_IPV6
-
 int get_pasv_socket_ipv6(struct connection *c, int cc, int *sock, unsigned char *result)
 {
 	int s;
@@ -277,7 +294,7 @@ int get_pasv_socket_ipv6(struct connection *c, int cc, int *sock, unsigned char
 		errno = EINVAL;
 		goto e;
 	}
-	EINTRLOOP(s, socket(PF_INET6, SOCK_STREAM, IPPROTO_TCP));
+	s = c_socket(PF_INET6, SOCK_STREAM, IPPROTO_TCP);
 	if (s == -1) goto e;
 	*sock = s;
 	set_nonblock(s);
@@ -315,37 +332,83 @@ int get_pasv_socket_ipv6(struct connection *c, int cc, int *sock, unsigned char
 	retry_connection(c);
 	return -1;
 }
-
 #endif
 
 #ifdef HAVE_SSL
-static void ssl_want_read(struct connection *c)
+static void ssl_setup_downgrade(struct connection *c)
+{
+#if !defined(HAVE_NSS)
+	int dd = c->no_tls;
+#ifdef SSL_OP_NO_TLSv1_2
+	if (dd) {
+		SSL_set_options(c->ssl, SSL_OP_NO_TLSv1_2);
+		dd--;
+	}
+#endif
+#ifdef SSL_OP_NO_TLSv1_1
+	if (dd) {
+		SSL_set_options(c->ssl, SSL_OP_NO_TLSv1_1);
+		dd--;
+	}
+#endif
+#if !defined(OPENSSL_NO_SSL2) || !defined(OPENSSL_NO_SSL3)
+	if (dd) {
+		SSL_set_options(c->ssl, SSL_OP_NO_TLSv1);
+		dd--;
+	}
+#endif
+#ifdef SSL_MODE_SEND_FALLBACK_SCSV
+	if (dd != c->no_tls) {
+		SSL_set_mode(c->ssl, SSL_MODE_SEND_FALLBACK_SCSV);
+	}
+#endif
+#endif
+}
+
+static void ssl_downgrade_dance(struct connection *c)
+{
+#if !defined(HAVE_NSS)
+	int downgrades = 0;
+#ifdef SSL_OP_NO_TLSv1_2
+	downgrades++;
+#endif
+#ifdef SSL_OP_NO_TLSv1_1
+	downgrades++;
+#endif
+#if !defined(OPENSSL_NO_SSL2) || !defined(OPENSSL_NO_SSL3)
+	downgrades++;
+#endif
+	if (++c->no_tls <= downgrades) {
+		retry_connect(c, S_SSL_ERROR, 1);
+	} else {
+		c->no_tls = 0;
+		retry_connect(c, S_SSL_ERROR, 0);
+	}
+#else
+	retry_connect(c, S_SSL_ERROR, 0);
+#endif
+}
+
+static void ssl_want_io(struct connection *c)
 {
 	int ret1, ret2;
 	struct conn_info *b = c->newconn;
 
 	set_connection_timeout(c);
 
-#ifndef HAVE_NSS
-	if (c->no_tsl) c->ssl->options |= SSL_OP_NO_TLSv1;
-#endif
 	switch ((ret2 = SSL_get_error(c->ssl, ret1 = SSL_connect(c->ssl)))) {
 		case SSL_ERROR_NONE:
-			c->newconn = NULL;
-			b->func(c);
-			mem_free(b);
+			connected_callback(c);
 			break;
 		case SSL_ERROR_WANT_READ:
-			set_handlers(*b->sock, (void(*)(void *))ssl_want_read, NULL, (void(*)(void *))exception, c);
+			set_handlers(*b->sock, (void(*)(void *))ssl_want_io, NULL, c);
 			break;
 		case SSL_ERROR_WANT_WRITE:
-			set_handlers(*b->sock, NULL, (void(*)(void *))ssl_want_read, (void(*)(void *))exception, c);
+			set_handlers(*b->sock, NULL, (void(*)(void *))ssl_want_io, c);
 			break;
 		default:
 			log_ssl_error(__LINE__, ret1, ret2);
-			c->no_tsl++;
-			setcstate(c, S_SSL_ERROR);
-			retry_connection(c);
+			ssl_downgrade_dance(c);
 			break;
 	}
 }
@@ -383,6 +446,10 @@ static void handle_socks(struct connection *c)
 		retry_connection(c);
 		return;
 	}
+	if (!b->socks_byte_count) {
+		log_data(command, len);
+		log_data(cast_uchar "\n", 1);
+	}
 	EINTRLOOP(wr, (int)write(*b->sock, command + b->socks_byte_count, len - b->socks_byte_count));
 	mem_free(command);
 	if (wr <= 0) {
@@ -392,11 +459,11 @@ static void handle_socks(struct connection *c)
 	}
 	b->socks_byte_count += wr;
 	if (b->socks_byte_count < len) {
-		set_handlers(*b->sock, NULL, (void(*)(void *))handle_socks, (void(*)(void *))exception, c);
+		set_handlers(*b->sock, NULL, (void(*)(void *))handle_socks, c);
 		return;
 	} else {
 		b->socks_byte_count = 0;
-		set_handlers(*b->sock, (void(*)(void *))handle_socks_reply, NULL, (void(*)(void *))exception, c);
+		set_handlers(*b->sock, (void(*)(void *))handle_socks_reply, NULL, c);
 		return;
 	}
 }
@@ -414,7 +481,7 @@ static void handle_socks_reply(struct connection *c)
 	}
 	b->socks_byte_count += rd;
 	if (b->socks_byte_count < (int)sizeof b->socks_reply) return;
-	/* debug("%x %x %x %x %x %x %x %x", b->socks_reply[0], b->socks_reply[1], b->socks_reply[2], b->socks_reply[3], b->socks_reply[4], b->socks_reply[5], b->socks_reply[6], b->socks_reply[7]); */
+	/*debug("%x %x %x %x %x %x %x %x", b->socks_reply[0], b->socks_reply[1], b->socks_reply[2], b->socks_reply[3], b->socks_reply[4], b->socks_reply[5], b->socks_reply[6], b->socks_reply[7]);*/
 	if (b->socks_reply[0]) {
 		setcstate(c, S_BAD_SOCKS_VERSION);
 		abort_connection(c);
@@ -454,12 +521,23 @@ static void dns_found(struct connection *c, int state)
 	try_connect(c);
 }
 
-static void retry_connect(struct connection *c, int err)
+static void retry_connect(struct connection *c, int err, int ssl_downgrade)
 {
 	struct conn_info *b = c->newconn;
 	if (!b->addr_index) b->first_error = err;
+#ifdef HAVE_SSL
+	if (c->ssl) {
+		if (c->ssl != DUMMY) SSL_free(c->ssl);
+		c->ssl = DUMMY;
+	}
+#endif
+	if (ssl_downgrade) {
+		close_socket(b->sock);
+		try_connect(c);
+		return;
+	}
 	b->addr_index++;
-	if (b->addr_index < b->addr.n) {
+	if (b->addr_index < b->addr.n && !b->dont_try_more_servers) {
 		close_socket(b->sock);
 		try_connect(c);
 	} else {
@@ -486,7 +564,7 @@ static void try_connect(struct connection *c)
 		return;
 	}
 	if (s == -1) {
-		retry_connect(c, get_error_from_errno(errno));
+		retry_connect(c, get_error_from_errno(errno), 0);
 		return;
 	}
 	set_nonblock(s);
@@ -519,10 +597,10 @@ static void try_connect(struct connection *c)
 #ifdef BEOS
 			if (errno == EWOULDBLOCK) errno = ETIMEDOUT;
 #endif
-			retry_connect(c, get_error_from_errno(errno));
+			retry_connect(c, get_error_from_errno(errno), 0);
 			return;
 		}
-		set_handlers(s, NULL, (void(*)(void *))connected, (void(*)(void *))exception, c);
+		set_handlers(s, NULL, (void(*)(void *))connected, c);
 		setcstate(c, !b->addr_index ? S_CONN : S_CONN_ANOTHER);
 	} else {
 		connected(c);
@@ -556,7 +634,7 @@ static void connected(struct connection *c)
 		if (err >= 10000) err -= 10000;	/* Why does EMX return so large values? */
 	} else {
 		if (!(err = errno)) {
-			retry_connect(c, S_STATE);
+			retry_connect(c, S_STATE, 0);
 			return;
 		}
 	}
@@ -568,53 +646,128 @@ static void connected(struct connection *c)
 #ifdef DOS
 		if (err == EALREADY) err = ETIMEDOUT;
 #endif
-		retry_connect(c, get_error_from_errno(err));
+		retry_connect(c, get_error_from_errno(err), 0);
 		return;
 	}
 #endif
-	if (b->addr_index) {
-		int i;
-		for (i = 0; i < b->addr_index; i++)
-			dns_set_priority(b->host, &b->addr.a[i], 0);
-		dns_set_priority(b->host, &b->addr.a[i], 1);
-	}
 	set_connection_timeout(c);
 	if (b->real_port != -1) {
+		update_dns_priority(c);
 		handle_socks(c);
 		return;
 	}
 #ifdef HAVE_SSL
 	if (c->ssl) {
 		int ret1, ret2;
+		unsigned char *h = get_host_name(remove_proxy_prefix(c->url));
+		if (*h && h[strlen(cast_const_char h) - 1] == '.') {
+			h[strlen(cast_const_char h) - 1] = 0;
+		}
 		c->ssl = getSSL();
 		if (!c->ssl) {
+			ret1 = ret2 = 0;
+			mem_free(h);
 			goto ssl_error;
 		}
-		SSL_set_fd(c->ssl, *b->sock);
-#ifndef HAVE_NSS
-		if (c->no_tsl) c->ssl->options |= SSL_OP_NO_TLSv1;
+#if defined(HAVE_SSL_CERTIFICATES) && !defined(OPENSSL_NO_STDIO)
+		if (!proxies.only_proxies) {
+			if (ssl_options.client_cert_key[0]) {
+				SSL_use_PrivateKey_file(c->ssl, cast_const_char ssl_options.client_cert_key, SSL_FILETYPE_PEM);
+			}
+			if (ssl_options.client_cert_crt[0]) {
+				SSL_use_certificate_file(c->ssl, cast_const_char ssl_options.client_cert_crt, SSL_FILETYPE_PEM);
+			}
+		}
 #endif
+		SSL_set_fd(c->ssl, *b->sock);
+		ssl_setup_downgrade(c);
+#ifdef SSL_CTRL_SET_TLSEXT_HOSTNAME
+		if (h[0] == '[' || !numeric_ip_address(h, NULL)
+#ifdef SUPPORT_IPV6
+		    || !numeric_ipv6_address(h, NULL, NULL)
+#endif
+		    ) goto skip_numeric_address;
+		SSL_set_tlsext_host_name(c->ssl, h);
+skip_numeric_address:
+#endif
+		mem_free(h);
 		switch ((ret2 = SSL_get_error(c->ssl, ret1 = SSL_connect(c->ssl)))) {
 			case SSL_ERROR_WANT_READ:
 				setcstate(c, S_SSL_NEG);
-				set_handlers(*b->sock, (void(*)(void *))ssl_want_read, NULL, (void(*)(void *))exception, c);
+				set_handlers(*b->sock, (void(*)(void *))ssl_want_io, NULL, c);
 				return;
 			case SSL_ERROR_WANT_WRITE:
 				setcstate(c, S_SSL_NEG);
-				set_handlers(*b->sock, NULL, (void(*)(void *))ssl_want_read, (void(*)(void *))exception, c);
+				set_handlers(*b->sock, NULL, (void(*)(void *))ssl_want_io, c);
 				return;
 			case SSL_ERROR_NONE:
 				break;
 			default:
-				log_ssl_error(__LINE__, ret1, ret2);
 			ssl_error:
-				c->no_tsl++;
-				setcstate(c, S_SSL_ERROR);
-				retry_connection(c);
+				log_ssl_error(__LINE__, ret1, ret2);
+				ssl_downgrade_dance(c);
 				return;
 		}
 	}
 #endif
+	connected_callback(c);
+}
+
+static void update_dns_priority(struct connection *c)
+{
+	struct conn_info *b = c->newconn;
+	if (!b->dont_try_more_servers) {
+		if (b->addr_index) {
+			int i;
+			for (i = 0; i < b->addr_index; i++)
+				dns_set_priority(b->host, &b->addr.a[i], 0);
+			dns_set_priority(b->host, &b->addr.a[i], 1);
+			b->addr_index = 0;
+		}
+		b->dont_try_more_servers = 1;
+	}
+}
+
+static void connected_callback(struct connection *c)
+{
+	struct conn_info *b = c->newconn;
+#ifdef HAVE_SSL_CERTIFICATES
+	if (c->ssl) {
+		if (ssl_options.certificates != SSL_ACCEPT_INVALID_CERTIFICATE) {
+			unsigned char *h = get_host_name(remove_proxy_prefix(c->url));
+			int err = verify_ssl_certificate(c->ssl, h);
+			if (err) {
+				if (ssl_options.certificates == SSL_WARN_ON_INVALID_CERTIFICATE) {
+					int flags = get_blacklist_flags(h);
+					if (flags & BL_IGNORE_CERTIFICATE)
+						goto ignore_cert;
+				}
+				mem_free(h);
+				setcstate(c, err);
+				abort_connection(c);
+				return;
+			}
+			ignore_cert:
+
+			err = verify_ssl_cipher(c->ssl);
+			if (err) {
+				if (ssl_options.certificates == SSL_WARN_ON_INVALID_CERTIFICATE) {
+					int flags = get_blacklist_flags(h);
+					if (flags & BL_IGNORE_CIPHER)
+						goto ignore_cipher;
+				}
+				mem_free(h);
+				setcstate(c, err);
+				abort_connection(c);
+				return;
+			}
+			ignore_cipher:
+
+			mem_free(h);
+		}
+	}
+#endif
+	update_dns_priority(c);
 	c->newconn = NULL;
 	b->func(c);
 	mem_free(b);
@@ -644,7 +797,7 @@ static void write_select(struct connection *c)
 	printf("-\n");*/
 
 #ifdef HAVE_SSL
-	if(c->ssl) {
+	if (c->ssl) {
 		if ((wr = SSL_write(c->ssl, wb->data + wb->pos, wb->len - wb->pos)) <= 0) {
 			int err;
 			if ((err = SSL_get_error(c->ssl, wr)) != SSL_ERROR_WANT_WRITE) {
@@ -677,7 +830,7 @@ static void write_select(struct connection *c)
 	if ((wb->pos += wr) == wb->len) {
 		void (*f)(struct connection *) = wb->done;
 		c->buffer = NULL;
-		set_handlers(wb->sock, NULL, NULL, NULL, NULL);
+		set_handlers(wb->sock, NULL, NULL, NULL);
 		mem_free(wb);
 		f(c);
 	}
@@ -696,13 +849,15 @@ void write_to_socket(struct connection *c, int s, unsigned char *data, int len,
 	memcpy(wb->data, data, len);
 	if (c->buffer) mem_free(c->buffer);
 	c->buffer = wb;
-	set_handlers(s, NULL, (void (*)(void *))write_select, (void (*)(void *))exception, c);
+	set_handlers(s, NULL, (void (*)(void *))write_select, c);
 }
 
 #define READ_SIZE	64240
+#define TOTAL_READ	(4193008 - READ_SIZE)
 
 static void read_select(struct connection *c)
 {
+	int total_read = 0;
 	struct read_buffer *rb;
 	int rd;
 	if (!(rb = c->buffer)) {
@@ -711,15 +866,18 @@ static void read_select(struct connection *c)
 		abort_connection(c);
 		return;
 	}
-	set_handlers(rb->sock, NULL, NULL, NULL, NULL);
+	set_handlers(rb->sock, NULL, NULL, NULL);
+
+read_more:
 	if ((unsigned)rb->len > MAXINT - sizeof(struct read_buffer) - READ_SIZE) overalloc();
 	rb = mem_realloc(rb, sizeof(struct read_buffer) + rb->len + READ_SIZE);
 	c->buffer = rb;
 
 #ifdef HAVE_SSL
-	if(c->ssl) {
+	if (c->ssl) {
 		if ((rd = SSL_read(c->ssl, rb->data + rb->len, READ_SIZE)) <= 0) {
 			int err;
+			if (total_read) goto success;
 			if ((err = SSL_get_error(c->ssl, rd)) == SSL_ERROR_WANT_READ) {
 				read_from_socket(c, rb->sock, rb, rb->done);
 				return;
@@ -742,6 +900,7 @@ static void read_select(struct connection *c)
 	{
 		EINTRLOOP(rd, (int)read(rb->sock, rb->data + rb->len, READ_SIZE));
 		if (rd <= 0) {
+			if (total_read) goto success;
 			if (rb->close && !rd) {
 				rb->close = 2;
 				rb->done(c, rb);
@@ -771,6 +930,17 @@ static void read_select(struct connection *c)
 	}
 	log_data(rb->data + rb->len, rd);
 	rb->len += rd;
+	total_read += rd;
+
+	if ((rd == READ_SIZE
+#ifdef HAVE_SSL
+	    || c->ssl
+#endif
+	    ) && total_read <= TOTAL_READ) {
+		if (can_read(rb->sock))
+			goto read_more;
+	}
+success:
 	rb->done(c, rb);
 }
 
@@ -788,7 +958,7 @@ void read_from_socket(struct connection *c, int s, struct read_buffer *buf, void
 	buf->sock = s;
 	if (c->buffer && buf != c->buffer) mem_free(c->buffer);
 	c->buffer = buf;
-	set_handlers(s, (void (*)(void *))read_select, NULL, (void (*)(void *))exception, c);
+	set_handlers(s, (void (*)(void *))read_select, NULL, c);
 }
 
 void kill_buffer_data(struct read_buffer *rb, int n)
