@@ -8,8 +8,8 @@
 
 #ifdef GRDRV_FB
 
-#define USE_GPM_DX
 /*#define USE_FB_ACCEL*/
+/*#define USE_FB_ACCEL_FILLRECT*/
 
 /* #define FB_DEBUG */
 /* #define SC_DEBUG */
@@ -92,13 +92,6 @@ struct fb_fillrect {
 
 static int TTY = 0;
 
-#ifndef USE_GPM_DX
-static int fb_txt_xsize, fb_txt_ysize;
-static struct winsize fb_old_ws;
-static struct winsize fb_new_ws;
-static int fb_old_ws_v;
-static int fb_msetsize;
-#endif
 static int fb_hgpm;
 
 static int fb_console;
@@ -192,8 +185,7 @@ static int last_mouse_buttons;
 		INC_IN_GR \
 		if (!fb_active) { END_GR return; }
 #define START_GR_0 \
-		in_gr_operation++; \
-		do_not_optimize_here(&fb_vmem); \
+		INC_IN_GR \
 		if (!fb_active) { END_GR return 0; }
 
 
@@ -218,7 +210,7 @@ static int last_mouse_buttons;
 	int xs=hndl->x,ys=hndl->y;\
 	unsigned char *data=hndl->data;\
 \
- 	TEST_INACTIVITY\
+	TEST_INACTIVITY\
 	if (x>=dev->clip.x2||x+xs<=dev->clip.x1) return;\
 	if (y>=dev->clip.y2||y+ys<=dev->clip.y1) return;\
 	if (x+xs>dev->clip.x2) xs=dev->clip.x2-x;\
@@ -301,7 +293,7 @@ static void redraw_mouse(void);
 
 static void fb_mouse_move(int dx, int dy)
 {
-	struct event ev;
+	struct links_event ev;
 	mouse_x += dx;
 	mouse_y += dy;
 	ev.ev = EV_MOUSE;
@@ -323,8 +315,8 @@ static void fb_mouse_move(int dx, int dy)
 
 static void fb_key_in(void *p, unsigned char *ev_, int size)
 {
-	struct event *ev = (struct event *)(void *)ev_;
-	if (size != sizeof(struct event)) return;
+	struct links_event *ev = (struct links_event *)(void *)ev_;
+	if (size != sizeof(struct links_event)) return;
 	if (ev->ev == EV_ABORT) terminate_loop = 1;
 	if (ev->ev != EV_KBD) return;
 	if ((ev->y & KBD_ALT) && ev->x >= '0' && ev->x <= '9') {
@@ -336,10 +328,9 @@ static void fb_key_in(void *p, unsigned char *ev_, int size)
 	else if (!ev->y && ev->x == KBD_F6) fb_mouse_move(0, 3);
 	else if (!ev->y && ev->x == KBD_F7) fb_mouse_move(0, -3);
 	else if (!ev->y && ev->x == KBD_F8) fb_mouse_move(3, 0);
-	else 
-	{
-		if (fb_driver.codepage!=utf8_table&&(ev->x)>=128&&(ev->x)<=255)
-			if ((ev->x=cp2u(ev->x,fb_driver.codepage)) == -1) return;
+	else {
+		if (g_kbd_codepage(&fb_driver) != utf8_table && ev->x >= 128 && ev->x <= 255)
+			if ((ev->x = cp2u(ev->x, g_kbd_codepage(&fb_driver))) == -1) return;
 		if (current_virtual_device->keyboard_handler) current_virtual_device->keyboard_handler(current_virtual_device, ev->x, ev->y);
 	}
 }
@@ -350,6 +341,16 @@ static void fb_key_in(void *p, unsigned char *ev_, int size)
 #define mouse_getscansegment(buf,x,y,w) memcpy(buf,fb_vmem+y*fb_linesize+x*fb_pixelsize,w)
 #define mouse_drawscansegment(ptr,x,y,w) memcpy(fb_vmem+y*fb_linesize+x*fb_pixelsize,ptr,w);
 
+#define do_with_mouse_device(cmd)				\
+do {								\
+	struct graphics_device *current_virtual_device_backup;	\
+								\
+	current_virtual_device_backup = current_virtual_device;	\
+	current_virtual_device = mouse_graphics_device;		\
+	cmd;							\
+	current_virtual_device = current_virtual_device_backup;	\
+} while (0)
+
 /* Flushes the background_buffer onscreen where it was originally taken from. */
 static void place_mouse_background(void)
 {
@@ -360,15 +361,7 @@ static void place_mouse_background(void)
 	bmp.skip=arrow_width*fb_pixelsize;
 	bmp.data=background_buffer;
 
-	{
-		struct graphics_device * current_virtual_device_backup;
-
-		current_virtual_device_backup=current_virtual_device;
-		current_virtual_device=mouse_graphics_device;
-		fb_draw_bitmap(mouse_graphics_device, &bmp, background_x,
-			background_y);
-		current_virtual_device=current_virtual_device_backup;
-	}
+	do_with_mouse_device(fb_draw_bitmap(mouse_graphics_device, &bmp, background_x, background_y));
 
 }
 
@@ -377,7 +370,6 @@ static void place_mouse_background(void)
  */
 static void hide_mouse(void)
 {
-	accel_sync();
 	global_mouse_hidden=1;
 	place_mouse_background();
 }
@@ -397,6 +389,8 @@ static void get_mouse_background(unsigned char *buffer_ptr)
 	width=fb_pixelsize*(arrow_width+x>fb_xsize?fb_xsize-x:arrow_width);
 	height=arrow_height+y>fb_ysize?fb_ysize-y:arrow_height;
 
+	accel_sync();
+
 	for (;height;height--){
 		mouse_getscansegment(buffer_ptr,x,y,width);
 		buffer_ptr+=skip;
@@ -411,7 +405,7 @@ static void render_mouse_arrow(void)
 {
 	int x,y, reg0, reg1;
 	unsigned char *mouse_ptr=mouse_buffer;
-	unsigned *arrow_ptr=arrow;
+	const unsigned *arrow_ptr=arrow;
 
 	for (y=arrow_height;y;y--){
 		reg0=*arrow_ptr;
@@ -438,14 +432,7 @@ static void place_mouse(void)
 	bmp.y=arrow_height;
 	bmp.skip=arrow_width*fb_pixelsize;
 	bmp.data=mouse_buffer;
-	{
-		struct graphics_device * current_graphics_device_backup;
-
-		current_graphics_device_backup=current_virtual_device;
-		current_virtual_device=mouse_graphics_device;
-		fb_draw_bitmap(mouse_graphics_device, &bmp, mouse_x, mouse_y);
-		current_virtual_device=current_graphics_device_backup;
-	}
+	do_with_mouse_device(fb_draw_bitmap(mouse_graphics_device, &bmp, mouse_x, mouse_y));
 	global_mouse_hidden=0;
 }
 
@@ -454,7 +441,6 @@ static void place_mouse(void)
  */
 static void show_mouse(void)
 {
-	accel_sync();
 	get_mouse_background(background_buffer);
 	background_x=mouse_x;
 	background_y=mouse_y;
@@ -519,6 +505,8 @@ static inline void place_mouse_composite(void)
 	if (mouse_bottom>fb_ysize) mouse_bottom=fb_ysize;
 	if (background_bottom>fb_ysize) background_bottom=fb_ysize;
 
+	accel_sync();
+
 	/* Let's do the top part */
 	if (background_top<mouse_top){
 		/* Draw the background */
@@ -569,7 +557,7 @@ static inline void place_mouse_composite(void)
 			if (background_length>0)
 				mouse_drawscansegment(
 					background_ptr +l2,
-				       	mouse_right,mouse_top ,l3);
+					mouse_right,mouse_top ,l3);
 			mouse_ptr+=skip;
 			background_ptr+=skip;
 		}
@@ -619,7 +607,6 @@ static inline void redraw_mouse_sophisticated(void)
 static void redraw_mouse(void)
 {
 	if (!fb_active) return; /* We are not drawing */
-	accel_sync();
 	if (mouse_x!=background_x||mouse_y!=background_y){
 		if (RECTANGLES_INTERSECT(
 			background_x, background_x+arrow_width,
@@ -647,7 +634,7 @@ static void redraw_mouse(void)
  * Good white purity
  * Correct rounding and dithering prediction
  * And this is the cabbala:
- * 063 021 063 
+ * 063 021 063
  * 009 009 021
  * 255 085 255
  * 036 036 084
@@ -659,26 +646,22 @@ static void generate_palette(struct palette *palette)
 	switch (fb_colors)
 	{
 		case 16:
-	       	for (a=0;a<fb_palette_colors;a++)
-		{
-       			palette->red[a]=(a&8)?65535:0;
-	       		palette->green[a]=((a>>1)&3)*(65535/3);
-		       	palette->blue[a]=(a&1)?65535:0;
-		}
-		break;
 		case 256:
-		for (a=0;a<fb_palette_colors;a++){
-		       	palette->red[a]=((a>>5)&7)*(65535./7.);
-			palette->green[a]=((a>>2)&7)*(65535./7.);
-       			palette->blue[a]=(a&3)*(65535/3);
+		for (a=0;a<fb_palette_colors;a++)
+		{
+			unsigned rgb[3];
+			q_palette(fb_colors, a, 65535, rgb);
+			palette->red[a] = rgb[0];
+			palette->green[a] = rgb[1];
+			palette->blue[a] = rgb[2];
 		}
 		break;
 		case 32768:
 		for (a=0;a<fb_palette_colors;a++){
 			/*
-		       	palette->red[a]=((a>>10)&31)*(65535/31);
+			palette->red[a]=((a>>10)&31)*(65535/31);
 			palette->green[a]=((a>>5)&31)*(65535/31);
-       			palette->blue[a]=(a&31)*(65535/31);
+			palette->blue[a]=(a&31)*(65535/31);
 			*/
 			palette->red[a]=
 			palette->green[a]=
@@ -688,9 +671,9 @@ static void generate_palette(struct palette *palette)
 		case 65536:
 		for (a=0;a<fb_palette_colors;a++){
 			/*
-		       	palette->red[a]=((a>>11)&31)*(65535/31);
+			palette->red[a]=((a>>11)&31)*(65535/31);
 			palette->green[a]=((a>>5)&63)*(65535/63);
-       			palette->blue[a]=(a&31)*(65535/31);
+			palette->blue[a]=(a&31)*(65535/31);
 			*/
 			palette->green[a]=(((a&63)*255)/64)*257;
 			palette->red[a]=
@@ -815,7 +798,7 @@ static void fb_clear_videoram(void)
 {
 	int rs;
 	START_GR
-#ifdef USE_FB_ACCEL
+#ifdef USE_FB_ACCEL_FILLRECT
 	if (accel_flags & FB_ACCEL_FILLRECT_ACCELERATED) {
 		struct fb_fillrect f;
 		f.dx = 0;
@@ -943,7 +926,8 @@ static void fb_shutdown_palette(void)
 {
 	if (have_cmap)
 	{
-		set_palette(&old_palette);
+		if (fb_active)
+			set_palette(&old_palette);
 		free_palette(&old_palette);
 		free_palette(&global_pal);
 	}
@@ -954,54 +938,18 @@ static void fb_ctrl_c(struct itrm *i)
 	kbd_ctrl_c();
 }
 
-#ifndef USE_GPM_DX
-static void fb_mouse_setsize(void)
-{
-	struct vt_stat vs;
-	int rs;
-	EINTRLOOP(rs, ioctl(0, VT_GETSTATE, &vs));
-	if (!rs) {
-		fd_set zero;
-		struct timeval tv;
-		FD_ZERO(&zero);
-		memset(&tv, 0, sizeof tv);
-		EINTRLOOP(rs, ioctl(0, VT_ACTIVATE, vs.v_active > 1 ? 1 : 2));
-		tv.tv_sec = 0;
-		tv.tv_usec = 100000;
-		EINTRLOOP(rs, select(0, &zero, &zero, &zero, &tv));
-		tv.tv_sec = 0;
-		tv.tv_usec = 100000;
-		EINTRLOOP(rs, select(0, &zero, &zero, &zero, &tv));
-		tv.tv_sec = 0;
-		tv.tv_usec = 100000;
-		EINTRLOOP(rs, select(0, &zero, &zero, &zero, &tv));
-		EINTRLOOP(rs, ioctl(0, VT_ACTIVATE, vs.v_active));
-	}
-}
-#endif
-
 static void unhandle_fb_mouse(void);
 
 static void fb_gpm_in(void *nic)
 {
-#ifndef USE_GPM_DX
-	static int lx = -1, ly = -1;
-#endif
-	struct event ev;
+	struct links_event ev;
 	Gpm_Event gev;
 	again:
 	if (Gpm_GetEvent(&gev) <= 0) {
 		unhandle_fb_mouse();
 		return;
 	}
-	/*fprintf(stderr, "%d %d\n", gev.x, gev.y);*/
-#ifndef USE_GPM_DX
-	if (gev.x != lx || gev.y != ly) {
-		mouse_x = (gev.x - 1) * fb_xsize / fb_txt_xsize + fb_xsize / fb_txt_xsize / 2 - 1;
-		mouse_y = (gev.y - 1) * fb_ysize / fb_txt_ysize + fb_ysize / fb_txt_ysize / 2 - 1;
-		lx = gev.x, ly = gev.y;
-	}
-#else
+	/*fprintf(stderr, "%x %x %d %d %d %d\n", gev.type, gev.buttons, gev.dx, gev.dy, gev.wdx, gev.wdy);*/
 	if (gev.dx || gev.dy) {
 		if (!((int)gev.type & gpm_smooth)) {
 			mouse_x += gev.dx * 8;
@@ -1014,7 +962,6 @@ static void fb_gpm_in(void *nic)
 		}
 #endif
 	}
-#endif
 	ev.ev = EV_MOUSE;
 	if (mouse_x >= fb_xsize) mouse_x = fb_xsize - 1;
 	if (mouse_y >= fb_ysize) mouse_y = fb_ysize - 1;
@@ -1041,13 +988,24 @@ static void fb_gpm_in(void *nic)
 	else if ((int)gev.type & GPM_DRAG) ev.b |= B_DRAG;
 	else ev.b |= B_MOVE;
 
-#ifndef USE_GPM_DX
-	if (fb_msetsize < 0) {
-	} else if (fb_msetsize < 10) {
-		fb_msetsize++;
-	} else if ((ev.b & BM_ACT) == B_MOVE && !(ev.b & BM_BUTT)) {
-		fb_mouse_setsize();
-		fb_msetsize = -1;
+#ifdef HAVE_WDX_WDY
+	if ((ev.b & BM_ACT) == B_DRAG || (ev.b & BM_ACT) == B_MOVE) {
+		if (gev.wdy < 0) {
+			ev.b &= ~BM_BUTT;
+			ev.b |= B_WHEELDOWN;
+		} else if (gev.wdy > 0) {
+			ev.b &= ~BM_BUTT;
+			ev.b |= B_WHEELUP;
+		}
+#if 0
+	/* it doesn't work anyway - the exps2 protocol doesn't support it and evdev support in gpm is buggy */
+		else if (gev.wdx < 0) {
+			ev.b &= ~BM_BUTT;
+			ev.b |= B_WHEELRIGHT;
+		} else if (gev.wdx > 0) {
+			ev.b &= ~BM_BUTT;
+			ev.b |= B_WHEELLEFT;
+#endif
 	}
 #endif
 
@@ -1064,29 +1022,7 @@ static void fb_gpm_in(void *nic)
 static int handle_fb_mouse(void)
 {
 	Gpm_Connect conn;
-#ifndef USE_GPM_DX
-	int gpm_ver = 0;
-	struct winsize ws;
-	fb_old_ws_v = 0;
-#endif
 	fb_hgpm = -1;
-#ifndef USE_GPM_DX
-	Gpm_GetLibVersion(&gpm_ver);
-	fb_msetsize = -1;
-	if (gpm_ver >= 11900) {
-		int rs;
-		EINTRLOOP(rs,ioctl(1, TIOCGWINSZ, &ws));
-		if (rs != -1) {
-			memcpy(&fb_old_ws, &ws, sizeof(struct winsize));
-			fb_old_ws_v = 1;
-			ws.ws_row *= 2;
-			EINTRLOOP(rs, ioctl(1, TIOCSWINSZ, &ws));
-			fb_msetsize = 0;
-			memcpy(&fb_new_ws, &ws, sizeof ws);
-		}
-	}
-	get_terminal_size(1, &fb_txt_xsize, &fb_txt_ysize);
-#endif
 	conn.eventMask = (unsigned short)~0U;
 	conn.defaultMask = gpm_smooth;
 	conn.minMod = 0;
@@ -1095,7 +1031,7 @@ static int handle_fb_mouse(void)
 		unhandle_fb_mouse();
 		return -1;
 	}
-	set_handlers(fb_hgpm, fb_gpm_in, (void (*)(void *))NULL, (void (*)(void *))NULL, NULL);
+	set_handlers(fb_hgpm, fb_gpm_in, (void (*)(void *))NULL, NULL);
 #ifdef SIGTSTP
 	install_signal_handler(SIGTSTP, (void (*)(void *))sig_tstp, NULL, 0);
 #endif
@@ -1111,15 +1047,7 @@ static int handle_fb_mouse(void)
 
 static void unhandle_fb_mouse(void)
 {
-	if (fb_hgpm >= 0) set_handlers(fb_hgpm, (void (*)(void *))NULL, (void (*)(void *))NULL, (void (*)(void *))NULL, NULL);
-#ifndef USE_GPM_DX
-	fb_hgpm = -1;
-	if (fb_old_ws_v) {
-		int rs;
-		EINTRLOOP(rs, ioctl(1, TIOCSWINSZ, &fb_old_ws));
-		fb_old_ws_v = 0;
-	}
-#endif
+	if (fb_hgpm >= 0) set_handlers(fb_hgpm, (void (*)(void *))NULL, (void (*)(void *))NULL, NULL);
 	Gpm_Close();
 #ifdef SIGTSTP
 	install_signal_handler(SIGTSTP, (void (*)(void *))sig_tstp, NULL, 0);
@@ -1245,7 +1173,7 @@ static unsigned char *fb_init_driver(unsigned char *param, unsigned char *ignore
 		return e;
 	}
 
-	EINTRLOOP(fb_handle, open("/dev/fb0", O_RDWR));
+	fb_handle = c_open(cast_uchar "/dev/fb0", O_RDWR | O_NOCTTY);
 	if (fb_handle==-1) {
 		fb_switch_shutdown();
 		if(fb_driver_param) { mem_free(fb_driver_param); fb_driver_param=NULL; }
@@ -1518,12 +1446,87 @@ static void fb_shutdown_driver(void)
 	if (in_gr_operation) internal("fb_shutdown_driver: in_gr_operation %d", in_gr_operation);
 }
 
+static void fb_after_fork(void)
+{
+	int rs;
+	EINTRLOOP(rs, close(fb_handle));
+}
 
 static unsigned char *fb_get_driver_param(void)
 {
 	return fb_driver_param;
 }
 
+static void fb_get_margin(int *left, int *right, int *top, int *bottom)
+{
+	*left = border_left;
+	*right = border_right;
+	*top = border_top;
+	*bottom = border_bottom;
+}
+
+static int fb_set_margin(int left, int right, int top, int bottom)
+{
+	struct rect_set *rs;
+	int i, l;
+	struct rect r;
+
+	if (left + right >= fb_xsize + border_left + border_right)
+		return -1;
+	if (top + bottom >= fb_ysize + border_top + border_bottom)
+		return -1;
+
+	hide_mouse();
+
+	rs = init_rect_set();
+	add_to_rect_set(&rs, &mouse_graphics_device->size);
+	r.x1 = left - border_left;
+	r.y1 = top - border_top;
+	r.x2 = fb_xsize - (right - border_right);
+	r.y2 = fb_ysize - (bottom - border_bottom);
+	exclude_rect_from_set(&rs, &r);
+	do_with_mouse_device(
+	for (i = 0; i < rs->m; i++) if (is_rect_valid(&rs->r[i]))
+		drv->fill_area(current_virtual_device, rs->r[i].x1, rs->r[i].y1, rs->r[i].x2, rs->r[i].y2, 0);
+	);
+	mem_free(rs);
+
+	mouse_x += border_left - left;
+	mouse_y += border_top - top;
+	fb_xsize = fb_xsize + border_left + border_right - (left + right);
+	fb_ysize = fb_ysize + border_top + border_bottom - (top + bottom);
+	border_left = left;
+	border_right = right;
+	border_top = top;
+	border_bottom = bottom;
+
+	if (mouse_x >= fb_xsize) mouse_x = fb_xsize - 1;
+	if (mouse_x < 0) mouse_x = 0;
+	if (mouse_y >= fb_ysize) mouse_y = fb_ysize - 1;
+	if (mouse_y < 0) mouse_y = 0;
+
+	fb_vmem = fb_mem + border_left * fb_pixelsize + border_top * fb_linesize;
+
+	mouse_graphics_device->size.x2 = mouse_graphics_device->clip.x2 = fb_xsize;
+	mouse_graphics_device->size.y2 = mouse_graphics_device->clip.y2 = fb_ysize;
+
+	show_mouse();
+
+	if (fb_driver_param) mem_free(fb_driver_param);
+	fb_driver_param = init_str();
+	l = 0;
+	add_num_to_str(&fb_driver_param, &l, left);
+	add_chr_to_str(&fb_driver_param, &l, ',');
+	add_num_to_str(&fb_driver_param, &l, top);
+	add_chr_to_str(&fb_driver_param, &l, ',');
+	add_num_to_str(&fb_driver_param, &l, right);
+	add_chr_to_str(&fb_driver_param, &l, ',');
+	add_num_to_str(&fb_driver_param, &l, bottom);
+
+	resize_virtual_devices(fb_xsize, fb_ysize);
+
+	return 0;
+}
 
 /* Return value:        0 alloced on heap
  *                      1 alloced in vidram
@@ -1579,7 +1582,7 @@ static void fb_draw_bitmap(struct graphics_device *dev,struct bitmap* hndl, int
 }
 
 
-#ifdef USE_FB_ACCEL
+#ifdef USE_FB_ACCEL_FILLRECT
 static inline unsigned fb_accel_color(void *ptr)
 {
 	return *(unsigned char *)ptr;
@@ -1595,7 +1598,7 @@ static void fb_fill_area(struct graphics_device *dev, int left, int top, int rig
 
 	FILL_CLIP_PREFACE
 
-#ifdef USE_FB_ACCEL
+#ifdef USE_FB_ACCEL_FILLRECT
 	if (accel_flags & FB_ACCEL_FILLRECT_ACCELERATED) {
 		struct fb_fillrect f;
 		f.dx = left + border_left;
@@ -1634,7 +1637,7 @@ static void fb_draw_hline(struct graphics_device *dev, int left, int y, int righ
 
 	HLINE_CLIP_PREFACE
 
-#ifdef USE_FB_ACCEL
+#ifdef USE_FB_ACCEL_FILLRECT
 	if (accel_flags & FB_ACCEL_FILLRECT_ACCELERATED) {
 		struct fb_fillrect f;
 		f.dx = left + border_left;
@@ -1671,7 +1674,7 @@ static void fb_draw_vline(struct graphics_device *dev, int x, int top, int botto
 
 	VLINE_CLIP_PREFACE
 
-#ifdef USE_FB_ACCEL
+#ifdef USE_FB_ACCEL_FILLRECT
 	if (accel_flags & FB_ACCEL_FILLRECT_ACCELERATED) {
 		struct fb_fillrect f;
 		f.dx = x + border_left;
@@ -1878,8 +1881,12 @@ struct graphics_driver fb_driver={
 	init_virtual_device,
 	shutdown_virtual_device,
 	fb_shutdown_driver,
-	dummy_emergency_shutdown,
+	(void (*)(void))NULL,
+	fb_after_fork,
 	fb_get_driver_param,
+	(unsigned char *(*)(void))NULL,
+	fb_get_margin,
+	fb_set_margin,
 	fb_get_empty_bitmap,
 	fb_register_bitmap,
 	fb_prepare_strip,
