@@ -6,6 +6,7 @@
 #include <errno.h>
 #include <stdlib.h>
 #include <fcntl.h>
+#include <string.h>
 #include <ctype.h>
 #include <png.h>
 
@@ -30,7 +31,7 @@ struct font{
 	unsigned char *adstyl;
 	unsigned char *spacing;
 	int n_letters; /* Number of letters in this font */
-	struct letter *letters; 
+	struct letter *letters;
 };
 
 struct font *fonts;
@@ -42,22 +43,39 @@ int file_select(const struct dirent * entry)
 {
 
 	const char *s=entry->d_name;
+	char *ext;
+	unsigned long int codepoint;
 	FILE *f;
 
-	if (strlen(s)!=8) return 0;
-	if (!((s[0]>='0'&&s[0]<='9')||(tolower(s[0])>='a'&&tolower(s[0])<='f'))) return 0;
-	if (!((s[1]>='0'&&s[1]<='9')||(tolower(s[1])>='a'&&tolower(s[1])<='f'))) return 0;
-	if (!((s[2]>='0'&&s[2]<='9')||(tolower(s[2])>='a'&&tolower(s[2])<='f'))) return 0;
-	if (!((s[3]>='0'&&s[3]<='9')||(tolower(s[3])>='a'&&tolower(s[3])<='f'))) return 0;
-	if (s[4]!='.'||tolower(s[5])!='p'||tolower(s[6])!='n'||tolower(s[7])!='g') return 0;
+	codepoint= strtoul(s, &ext, 16);
+	if (codepoint > 0x10FFFFUL || strcasecmp(ext, ".png"))
+		return 0;
 	f=fopen(s,"r");
 	if (!f) return 0; /* If it can't be open as a file then it's probably
 		           * a directory
-			   */ 
+			   */
 	fclose(f);
 	return 1;
 }
 
+int hexstrsort(const struct dirent **a, const struct dirent **b)
+{
+	const char *aname, *bname;
+	int diff;
+
+	for( aname= (**a).d_name; *aname == '0'; ++aname );
+	for( bname= (**b).d_name; *bname == '0'; ++bname );
+	diff= (int)strlen(aname) - (int)strlen(bname);
+	if( diff ) return diff;
+	while( *aname ) {
+		diff= (int)tolower(*aname) - (int)tolower(*bname);
+		if( diff ) return diff;
+		++aname;
+		++bname;
+	}
+	return 0;
+}
+
 /* If the directory name has bad structure (must have 4 dashes in the name),
  * then nothing is done
  * Fills weight, slant, adstyl, spacing with appropriate strings
@@ -239,7 +257,7 @@ two:
 				fprintf(output,"%c",c);
 				char_pos++;
 				forbidden_0_to_7=0;
-			
+
 			}
 			break;
 	}
@@ -247,8 +265,8 @@ two:
 }
 
 
-/* name:        path to file to process
- * output:      where to put the C code
+/* name:		path to file to process
+ * output:		where to put the C code
  * char_number: number of the char. -1 means do not write anything into the tables.
 */
 void process_file(unsigned char *name, FILE *output, int char_number)
@@ -257,7 +275,6 @@ void process_file(unsigned char *name, FILE *output, int char_number)
 	int c;
 	int letter_code=letter_code;
 	int count=0;
-	unsigned char btr[5];
 	int forbidden_0_to_7;
 
 	f=fopen((const char *)name,"r");
@@ -268,13 +285,11 @@ void process_file(unsigned char *name, FILE *output, int char_number)
 		exit(1);
 	}
 	if (char_number>=0){
-		memcpy(btr,name,4);
-		btr[4]=0;
-		letter_code=strtoul((const char *)btr,NULL,16);
+		letter_code=strtoul((const char *)name,NULL,16);
 		fonts[n_fonts-1].letters[char_number].begin=stamp;
 	}
 	char_pos = 38;
-	fprintf(output,"static unsigned char letter_%d[] = \"", stamp);
+	fprintf(output,"static const unsigned char letter_%d[] = \"", stamp);
 	forbidden_0_to_7=0;
 
 	while(EOF!=(c=fgetc(f))){
@@ -296,51 +311,59 @@ void process_letters(FILE *output)
 	struct dirent **namelist;
 	struct dirent **ptr;
 	int nr,a;
-	
-	nr=scandir(".",&namelist,file_select,alphasort);
+
+	nr=scandir(".",&namelist,file_select,hexstrsort);
 	if (nr<0){
 		perror(PROGNAME);
 		exit(1);
 	}
 	fonts[n_fonts-1].n_letters=nr;
 	if (!nr) return;
-	if (nr<0){
-		perror(PROGNAME);
-		exit(1);
-	}
 	n_letters+=nr;
 	fonts[n_fonts-1].letters=malloc(sizeof(struct letter)*nr);
 	if (!(fonts[n_fonts-1].letters)){
 		fprintf(stderr,"%s: Out of memory", PROGNAME);
 		exit(1);
 	}
-	ptr=namelist;                                    
+	ptr=namelist;
 	for(a=0;a<nr;a++){
 		process_file((unsigned char *)(*ptr)->d_name,output,a);
 		free(*ptr);
 		ptr++;
-	}	
+	}
 	free(namelist);
 }
 
+int dir_select(const struct dirent *directory_entry)
+{
+	struct stat stat0;
+	if (!strcmp(directory_entry->d_name, ".")) return 0;
+	if (!strcmp(directory_entry->d_name, "..")) return 0;
+	if (0>stat(directory_entry->d_name,&stat0)) return 0;
+		/* Can't be stated
+		 */
+	if (!S_ISDIR(stat0.st_mode)) return 0; /* Is no directory */
+	return 1;
+}
+
 /* build_font_table:
  * scans the font directory
  * for each directory:
- *  * renews font description structures
- *  * scans the images
- *  * builds a binary search array
- *  * writes the font structure into the ../font_inc.c
- *  * releases font description structures
- *  Also prints the font_data byte table.
+ *	* renews font description structures
+ *	* scans the images
+ *	* builds a binary search array
+ *	* writes the font structure into the ../font_inc.c
+ *	* releases font description structures
+ *	Also prints the font_data byte table.
  */
 void build_font_table(FILE *output)
 {
 	struct dirent *directory_entry;
 	unsigned char *directory_name;
-	DIR *dir;
-	struct stat stat0;
 	unsigned char *family, *weight, *slant, *adstyl, *spacing;
-	
+	struct dirent **namelist;
+	int j, nr;
+
 	directory_name=(unsigned char *)"system_font";
 	if (0>chdir((const char *)directory_name)){
 		fprintf(stderr,"%s: can't change into directory %s.\n",
@@ -354,28 +377,28 @@ void build_font_table(FILE *output)
 		perror(PROGNAME);
 		exit(1);
 	}
-	
+
 	directory_name=(unsigned char *)"font";
-	dir=opendir((const char *)directory_name);
 	if (0>chdir((const char *)directory_name)){
 		fprintf(stderr,"%s can't change into directory %s.\n",
 				PROGNAME,directory_name);
 		perror(PROGNAME);
 		exit(1);
 	}
-	if (!dir){
-		fprintf(stderr,"%s can't opendir() directory %s.\n",PROGNAME,
-				directory_name);
+	nr=scandir(".",&namelist,dir_select,alphasort);
+	if (nr < 0) {
+		fprintf(stderr, "scanlist failed\n");
 		perror(PROGNAME);
 		exit(1);
 	}
-	while((directory_entry=readdir(dir))){
+	if (!nr) {
+		fprintf(stderr, "no fonts\n");
+		exit(1);
+	}
+	for (j = 0; j < nr; j++) {
 		unsigned char *name;
 		int i;
-		if (0>stat(directory_entry->d_name,&stat0)) continue; 
-			/* Can't be stated
-			 */
-		if (!S_ISDIR(stat0.st_mode)) continue; /* Is no directory */
+		directory_entry = namelist[j];
 		/* Process the directory */
 		name = malloc(strlen(directory_entry->d_name) + 1);
 		if (!name){
@@ -385,6 +408,7 @@ void build_font_table(FILE *output)
 		for (i = 0; directory_entry->d_name[i]; i++)
 			name[i] = tolower(directory_entry->d_name[i]);
 		name[i] = 0;
+		free(directory_entry);
 		if (parse_font_name(name,&family,&weight,&slant,
 			&adstyl,&spacing))
 		{
@@ -406,6 +430,7 @@ void build_font_table(FILE *output)
 			exit(1);
 		}
 	}
+	free(namelist);
 	fprintf(output,"\n");
 }
 
@@ -485,7 +510,7 @@ int main(int argc, char **argv)
 		perror("font_inc.c");
 		exit(1);
 	}
-	
+
 	fprintf(output,"#include \"cfg.h\"\n\n");
 	fprintf(output,"#ifdef G\n\n");
 	fprintf(output,"#include \"links.h\"\n\n");
