@@ -23,14 +23,27 @@ struct jerr_struct{
 	jmp_buf setjmp_buffer;
 };
 
+struct jpg_decoder{
+	struct jpeg_decompress_struct *cinfo;
+	struct jerr_struct *jerr;
+	unsigned char state; /* 0: header 1: start 2: scanlines 3: end 4,5: also
+			something */
+	int skip_bytes;
+	unsigned char *jdata;
+	unsigned char *scanlines[16];
+};
+
 static struct jerr_struct *global_jerr;
 static struct jpeg_decompress_struct *global_cinfo;
-static int mesg_unsup_emitted; /* Defaults to zero at program startup and once
+static int mesg_unsup_emitted = 0; /* Defaults to zero at program startup and once
 				* set is never reset back to zero */
 
 
+/*#include <jerror.h>*/
+
 METHODDEF(void) my_error_exit(j_common_ptr cinfo)
 {
+	/*fprintf(stderr, "jpeg error %d (%d)\n", cinfo->err->msg_code, JERR_OUT_OF_MEMORY);*/
 	longjmp(global_jerr->setjmp_buffer,2);
 }
 
@@ -53,11 +66,11 @@ METHODDEF(void) my_skip_input_data(j_decompress_ptr cinfo,long num_bytes)
 {
 	if ((unsigned long)num_bytes>cinfo->src->bytes_in_buffer)
 	{
-	 	/* We have to enter skipping state */
-	 	cinfo->src->next_input_byte+=cinfo->src->bytes_in_buffer;
+		/* We have to enter skipping state */
+		cinfo->src->next_input_byte+=cinfo->src->bytes_in_buffer;
 		((struct jpg_decoder *)(global_cimg->decoder))->skip_bytes
-  			=num_bytes-cinfo->src->bytes_in_buffer;
-  		cinfo->src->bytes_in_buffer=0;
+			=(int)(num_bytes-cinfo->src->bytes_in_buffer);
+		cinfo->src->bytes_in_buffer=0;
 	}
 	else
 	{
@@ -87,7 +100,6 @@ g19_2000:
 	if (setjmp(global_jerr->setjmp_buffer)){
 		jpeg_destroy_decompress(global_cinfo);
 		goto g19_2000;
-		return;
 	}
 	jpeg_stdio_src(global_cinfo,stdin);
 	global_cinfo->src->init_source=&nop;
@@ -146,6 +158,9 @@ static void gray_to_rgb(unsigned char *data, int pixels)
 /* Fixes returned data in case they are CMYK or grayscale. */
 static inline void fix_data( struct jpg_decoder *deco, int lines_read)
 {
+#ifdef __ICC
+	volatile	/* ICC bug */
+#endif
 	int a;
 
 	switch (global_cinfo->output_components){
@@ -221,7 +236,7 @@ void jpeg_restart(struct cached_image *cimg, unsigned char *data, int length)
 
 	/* Update the next input byte pointer for the decoder to continue at
 	 * the right position */
-	global_cinfo->src->next_input_byte=deco->jdata;
+	global_cinfo->src->next_input_byte=(void *)deco->jdata;
 
 	/* ...:::...:..:.:::.:.::::.::.:.:.:.::..::::.::::.:...: */
 	/* Update the length of data in the decoder buffer */
@@ -231,7 +246,7 @@ void jpeg_restart(struct cached_image *cimg, unsigned char *data, int length)
 	switch(deco->state){
 		case 0:
 		/* jpeg_read_header */
-	   	if (JPEG_SUSPENDED==jpeg_read_header(global_cinfo,TRUE))
+		if (JPEG_SUSPENDED==jpeg_read_header(global_cinfo,TRUE))
 			break;
 		global_cinfo->buffered_image=TRUE;
 		deco->state=1;
@@ -275,7 +290,7 @@ void jpeg_restart(struct cached_image *cimg, unsigned char *data, int length)
 			/* Let's make a decompression fatal error here */
 
 			if (!mesg_unsup_emitted){
-				fprintf(stderr,
+				error(
 			"Unsupported JPEG output components number: %d.\n",
 			cimg->buffer_bytes_per_pixel);
 				mesg_unsup_emitted=1;
@@ -286,17 +301,17 @@ void jpeg_restart(struct cached_image *cimg, unsigned char *data, int length)
 			 * understand and maintain.  If possible an alternative
 			 * should be used. Hahaha :) ;-)
 			 */
-			/* Free will makes people hard to understand 
-			 * and maintain. If possible an alternative should be 
+			/* Free will makes people hard to understand
+			 * and maintain. If possible an alternative should be
 			 * used.
 			 */
 			/* With our new LongJump(TM) your jumps will be longer
 			 * than with ordinary commercially available jumps.
 			 */
 		}
-		cimg->red_gamma=sRGB_gamma;
-		cimg->green_gamma=sRGB_gamma;
-		cimg->blue_gamma=sRGB_gamma;
+		cimg->red_gamma=(float)sRGB_gamma;
+		cimg->green_gamma=(float)sRGB_gamma;
+		cimg->blue_gamma=(float)sRGB_gamma;
 		/* This is defined in the JPEG standard somehow that sRGB
 		 * color space is used. */
 
@@ -325,7 +340,7 @@ susp0:
 		case 3:
 		/* jpeg_read_scanlines */
 			/* color */
-        	while (global_cinfo->output_scanline
+		while (global_cinfo->output_scanline
 			<global_cinfo->output_height){
 			int a, lines;
 
@@ -336,19 +351,19 @@ susp0:
 				*global_cinfo->output_width*cimg->
 					buffer_bytes_per_pixel;
 			}
-		
-         		if ((lines=
+
+			if ((lines=
 				jpeg_read_scanlines(
 				global_cinfo,deco->scanlines,1))){
 				/* Some lines were written into cimg buffer */
 				cimg->rows_added=1;
-				
+
 				fix_data(deco, lines);
 			}else{
 				/* No lines have been written into cimg
 				 * buffer */
-                	 	/* We are suspended and we want more data */
-         	 		goto susp0; /* Break the outer 
+				/* We are suspended and we want more data */
+				goto susp0; /* Break the outer
 					     * switch statement */
 			}
 		}
@@ -377,19 +392,28 @@ decoder_ended:
 	}
 }
 
+void jpeg_destroy_decoder(struct cached_image *cimg)
+{
+	struct jpg_decoder *deco = (struct jpg_decoder *)cimg->decoder;
+	jpeg_destroy_decompress(deco->cinfo);
+	mem_free(deco->cinfo);
+	mem_free(deco->jerr);
+	if (deco->jdata) mem_free(deco->jdata);
+}
+
 void add_jpeg_version(unsigned char **s, int *l)
 {
-	add_to_str(s, l, "JPEG (");
+	add_to_str(s, l, cast_uchar "JPEG (");
 #if defined(JPEG_LIB_VERSION_MAJOR) && defined(JPEG_LIB_VERSION_MINOR)
 	add_num_to_str(s, l, JPEG_LIB_VERSION_MAJOR);
-	add_to_str(s, l, ".");
+	add_to_str(s, l, cast_uchar ".");
 	add_num_to_str(s, l, JPEG_LIB_VERSION_MINOR);
 #else
 	add_num_to_str(s, l, JPEG_LIB_VERSION / 10);
-	add_to_str(s, l, ".");
+	add_to_str(s, l, cast_uchar ".");
 	add_num_to_str(s, l, JPEG_LIB_VERSION % 10);
 #endif
-	add_to_str(s, l, ")");
+	add_to_str(s, l, cast_uchar ")");
 }
 
 #endif /* #ifdef HAVE_JPEG */
