@@ -31,23 +31,29 @@
 #define LEAK_DEBUG
 #endif
 
-#define LINKS_MANUAL_URL		(!strcmp(cast_const_char language_name(current_language),"Czech") || !strcmp(cast_const_char language_name(current_language),"Slovak")?\
+#define LINKS_MANUAL_URL		(!strcmp(cast_const_char language_name(get_current_language()),"Czech") || !strcmp(cast_const_char language_name(get_current_language()),"Slovak")?\
 					"http://links.twibright.com/user.html":\
 					"http://links.twibright.com/user_en.html")
-#define LINKS_HOMEPAGE_URL		(!strcmp(cast_const_char language_name(current_language),"Czech") || !strcmp(cast_const_char language_name(current_language),"Slovak")?\
+#define LINKS_HOMEPAGE_URL		(!strcmp(cast_const_char language_name(get_current_language()),"Czech") || !strcmp(cast_const_char language_name(get_current_language()),"Slovak")?\
 					"http://links.twibright.com/index_cz.php":\
 					"http://links.twibright.com/")
-#define LINKS_CALIBRATION_URL		(!strcmp(cast_const_char language_name(current_language),"Czech") || !strcmp(cast_const_char language_name(current_language),"Slovak")?\
+#define LINKS_CALIBRATION_URL		(!strcmp(cast_const_char language_name(get_current_language()),"Czech") || !strcmp(cast_const_char language_name(get_current_language()),"Slovak")?\
 					"http://links.twibright.com/kalibrace.html":\
 					"http://links.twibright.com/calibration.html")
 
 #define LINKS_SOCK_NAME			"socket"
 #define LINKS_PORT			23755
+#define LINKS_G_PORT_START		32768
+#define LINKS_G_PORT_LEN		16384
 #define MAX_BIND_TRIES			3
 #define AF_UNIX_SOCKET_TIMEOUT		2
 
 #define FG_POLL_TIME			500
 
+#define HEAPMIN_FACTOR			100
+#define HEAPMIN_INTERVAL		10000
+#define MALLOC_TRIM_INTERVAL		300000
+
 #define DNS_TIMEOUT			3600000UL
 
 #define HTTP_KEEPALIVE_TIMEOUT		300000
@@ -185,9 +191,13 @@
 
 #define MAX_INPUT_URL_LEN		4096
 
+#ifdef DOS
+#define SPD_DISP_TIME			1000
+#else
 #define SPD_DISP_TIME			200
-#define CURRENT_SPD_SEC			50
-#define CURRENT_SPD_AFTER		100
+#endif
+#define CURRENT_SPD_SEC			(10000 / SPD_DISP_TIME)
+#define CURRENT_SPD_AFTER		(20000 / SPD_DISP_TIME)
 
 #define RESOURCE_INFO_REFRESH		100
 
